/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.monitoring.core.bpel;


import java.util.Date;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.spagic.metadb.base.ProcessInstanceStateConstants;
import org.spagic.metadb.dbutils.AuditDBUtils;
import org.spagic.metadb.dbutils.HibernateUtil;
import org.spagic.metadb.model.LogJournal;
import org.spagic.metadb.model.ProcessInstance;
import org.spagic.metadb.model.RelevantData;
import org.spagic.metadb.model.Step;
import org.spagic.metadb.model.TransitionInstance;
import org.spagic.monitoring.core.MonitorServiceBase;
import org.spagic.monitoring.message.BPELMessage;
import org.spagic.monitoring.message.MessageIFace;
import org.spagic.utils.SpagicUtils;


/**
 * @author boccalon
 * Optimized BPEL monitor engine: avoids performing a lot of checks
 * necessary on JBI engines.
 */
public class MonitorServiceBPEL extends MonitorServiceBase {

	//	For a consumer endpoint, indicates if the processor
	//    should store the exchanges in the meta database, for
	//    using the restart feature.
	private boolean storeMessagesForRestart = true;
				
	public MonitorServiceBPEL(boolean storeMessagesForRestart,
						  Log log) {
		super();
		this.storeMessagesForRestart = storeMessagesForRestart;
		this.log = log;
	}

	/**
	 * Main method: perform the monitoring of all exchanges.
	 * @param exchange
	 */
	public void traceMessage(MessageIFace message) {
		
		Session aSession = null;
		Transaction tx = null;
		BPELMessage bpelMessage = (BPELMessage)message;
		try {
			TransitionInstance transitionInstance = null;
			aSession = HibernateUtil.getSessionFactory().openSession();
			tx = aSession.beginTransaction();

			debug("Retrieve the Communication Pattern");
			String communicationPattern = message.getCommunicationPattern();
			debug("Communication Pattern Retrieved ["+communicationPattern+"]");
			
			// If the state of the exchange is ACTIVE we doesn't search for the
			// exchange on the database because IT CANNOT exist: it is the first exchange
			// of the transition
			if (!message.isFirstMessageForTransition()) {
				debug("Search Transition Instance for a record with ID = " + message.getId());
				transitionInstance = AuditDBUtils.getTransitionInstanceByMessageId(aSession, message.getId());
				
				if (transitionInstance == null) {
					// Transition not found: should be the first transition of the process, the Receive activity
					debug("Transition Instance not found: first process transition");
					return;
				}
			}

			// Retrieve correlation id to discover the process to which this
			// exchange belongs to
			TransitionManagementState transitionManagementState = null;
			boolean processInstanceCreated = false;	// Setted to true if the instance of the service assembly
													// is created now: useful to limit the iter correlation work
													// only in this case
			
			if (bpelMessage.isProcessTerminated()) {
				// Retrieve exchange correlation id to discover the process to
				// which the exchange belongs to
				ProcessInstance processInstance = retrieveProcessInstance(aSession, message);
				// If the process is not registered for monitoring, we dont find it on database
				if (processInstance != null) {
					manageDone(processInstance, aSession);
				}
			} else {
				// Process running
				if (transitionInstance != null) {
					// There is already a Transition with this id logged: we
					// should update the existing row with the state of exchange
					debug(" Found Transition Instance -> Updating... " );
					transitionManagementState = manageExistingTransition(message, transitionInstance);
					debug(" Transition Instance Updated " );
				} else {
					debug("Creating a new Transition Instance.... " );
					TransitionInstanceCreationResult transitionInstanceCreationResult = createTransitionInstance(message, aSession);
					transitionInstance = transitionInstanceCreationResult.transitionInstance;
					processInstanceCreated = transitionInstanceCreationResult.processInstanceCreated;
					transitionManagementState = transitionInstanceCreationResult.transitionManagementState;
					debug((processInstanceCreated ? " Created a new Process Instance" : ""));
					
					if (transitionManagementState == null) {
						transitionManagementState = new TransitionManagementState();
					}
					
					if (transitionInstance != null) {
						transitionManagementState.transitionState = transitionInstance.getTransitionState();
					}
				}

				if (transitionInstance != null) {
					// Trace the message and relevant data
					aSession.save(transitionInstance);

					// Retrieve the message to trace
					if (message.existsMessageToTrace()) {
						try {
							LogJournal messageInfo = traceMessage(message, transitionInstance, transitionManagementState.transitionState, aSession);
							
							// Trace message relevant data
							List<RelevantData> savedRelevantData = traceRelevantData(messageInfo, aSession);
							
							// Check if the instance of service assembly was just created (and not saved yet)
							// Consider ONLY at least one relevant data have been produced
							if ((processInstanceCreated) && (savedRelevantData.size() > 0)) {
								checkPoolCorrelation(messageInfo, savedRelevantData, aSession);
							}
						} catch (Exception ex) {
				        	log.error(" Exception during tracing of message ", ex);
						}
					}
				}
				
				// Verify if the process is faulted
				// This step MUST be executed AFTER the checks about the transitions, to
				// allow proper update of transition state, so in the graph we can show
				// the component (or better, the transition) that produced the failure
				if (transitionManagementState.processFault) {
					manageFault(transitionInstance, aSession);
				}
				
			}
				
			aSession.flush();
			tx.commit();
		} catch (Exception e) {
        	log.error("traceExchange ", e);
			tx.rollback();
		} finally {
			if (aSession != null) {
				aSession.close();
			}
		}
	}
	
	private ProcessInstance retrieveProcessInstance(Session aSession, MessageIFace message) {
		// Retrieve exchange correlation id to discover the process to
		// which the exchange belongs to
		String correlationID = message.getCorrelationID();
		debug("createMessageExchange:: Correlation ID ["+correlationID+"]");
		ProcessInstance processInstance = AuditDBUtils.getProcessInstanceByIdCorrelation(aSession, correlationID);
		return processInstance;
	}
	
	private String getDBStepName(QName serviceName) {
		String namespaceURI = serviceName.getNamespaceURI();
		String localPart = serviceName.getLocalPart();
		
		// Extract process name and version from namespace
		int indexOfVersion = namespaceURI.lastIndexOf(":v");
		int indexOfProcessName = namespaceURI.indexOf(SpagicUtils.PREFIX_NAMESPACE) + SpagicUtils.PREFIX_NAMESPACE.length();
		String processName = namespaceURI.substring(indexOfProcessName, indexOfVersion);
		String version = "_v_" + namespaceURI.substring(indexOfVersion + 2);
		
		return processName + "." + localPart + version;
	}
	
	/**
	 * Create on the database a new TransitionInstance that corresponds to a new Message Exchange.
	 * @param message Current Message Exchange
	 * @param communicationPattern Message Exchange Pattern
	 * @param aSession Hibernate Session
	 * @return TransitionInstanceCreationResult result of the creation
	 */
	private TransitionInstanceCreationResult createTransitionInstance(MessageIFace message, Session aSession) {
		Step sourceStep = null;
		Step targetStep = null;

		// There isn't a Message Exchange with this id logged: we should
		// create a new one
		debug("createMessageExchange:: START");
		TransitionInstanceCreationResult result = new TransitionInstanceCreationResult();
		
		// Retrieve SOURCE service endpoint: service name (QName) + endpoint name
		QName sourceServiceQName = message.getSourceStepQName();
		if (sourceServiceQName == null) {
			// Not available in BPEL
	    	debug("createMessageExchange -> No configured Sender endpoint for exchange:" + message);
		} else {
			String sourceServiceName = sourceServiceQName.getLocalPart();
			sourceServiceName = sourceServiceName.substring(sourceServiceName.indexOf(':') + 1);
			sourceStep = AuditDBUtils.getStepByName(aSession, sourceServiceQName.getNamespaceURI(), sourceServiceName);
			
			if (sourceStep == null) {
				debug("Process not registered on database, sourceStep " + sourceStep + " not found");
				return result;
			} else if (!sourceStep.getProcess().isMonitorEnabled()) {
				debug("Monitoring on process " + sourceStep.getProcess().getName() + " not enabled");
				return result;
			}
		}

		// Retrieve TARGET service for the exchange
		QName targetServiceName = message.getTargetStepQName();
		targetStep = AuditDBUtils.getStepByName(aSession, targetServiceName.getNamespaceURI(), getDBStepName(targetServiceName));
		debug("createMessageExchange -> Target Step ["+targetStep+"]");

		if (targetStep == null) {
			debug("Process not registered on database, targetStep " + targetStep + " not found");
			debug("targetServiceName.getNamespaceURI(): " + targetServiceName.getNamespaceURI());
			debug("targetServiceName.getLocalPart(): " + targetServiceName.getLocalPart());
			return result;
		} else if (!targetStep.getProcess().isMonitorEnabled()) {
			// The step is replicated (there is the same check for sourceStep) because in some cases
			// the first check is not executed because the sourceServiceName is not available
			debug("Monitoring on process " + targetStep.getProcess().getName() + " not enabled");
			return result;
		}
		
		// Retrieve exchange correlation id to discover the process to
		// which the exchange belongs to
		ProcessInstance processInstance = retrieveProcessInstance(aSession, message);
				
		if (processInstance == null) {
			// This is a new process instance: create a new row
			// Before creating a new service assembly instance, I need to retrieve
			// the service assembly, and to retrieve it I can use the endpoint
			String correlationID = message.getCorrelationID();
			processInstance = createProcessInstance(targetStep.getProcess(), correlationID, aSession);
			result.processInstanceCreated = true;
		}

		// The instance of the process already exists:
		// we have to create the journal_instance_service row
		TransitionInstance transitionInstance = new TransitionInstance();
		transitionInstance.setProcessInstance(processInstance);
		transitionInstance.setIdMessage(message.getId());
		
		// Set the exchangeStart retrieving it from the MessageExchange's property
		transitionInstance.setTransitionStart(message.getTransitionStartDate());
		
		// Determine the state for this transition
		result.transitionManagementState = message.getStartState();
		transitionInstance.setTransitionState(result.transitionManagementState.transitionState);
		transitionInstance.setStepByIdStepSource(sourceStep);
		transitionInstance.setStepByIdStepTarget(targetStep);
		
		
		// Store Message Exchange for allow restart in case of errors
		if (storeMessagesForRestart) {
			transitionInstance.setTransitionData(message.getSerializedMessage());
		}
		
		result.transitionInstance = transitionInstance;
		return result;
	}
		
	/**
	 * Invoked when a FAULT message is received: it's responsible for updating the process state.
	 * @param transitionInstance Current TransitionInstance
	 * @param aSession Hibernate session
	 */
	private void manageFault(TransitionInstance transitionInstance, Session aSession) {
		ProcessInstance processInstance = transitionInstance.getProcessInstance();

		// The process has faulted: update its status
		processInstance.setState(ProcessInstanceStateConstants.PROCESS_FAULTED);
		Date endDate = new Date();
		Date startDate = processInstance.getStartdate();
		processInstance.setEnddate(endDate);
		processInstance.setExecutionTime(endDate.getTime() - startDate.getTime());
		/*Long idServiceAssemblySaved = (Long) */ aSession.save(processInstance);
	}
	
	/**
	 * Invoked when a DONE message is received: it's responsible for updating the process state.
	 * @param transitionInstance Current TransitionInstance
	 * @param aSession Hibernate session
	 */
	private void manageDone(ProcessInstance processInstance, Session aSession) {
		
		// The process should be finished: update its status
		processInstance.setState(ProcessInstanceStateConstants.PROCESS_DONE);
		Date endDate = new Date();
		Date startDate = processInstance.getStartdate();
		processInstance.setEnddate(endDate);
		processInstance.setExecutionTime(endDate.getTime() - startDate.getTime());
		/*Long idServiceAssemblySaved = (Long) */aSession.save(processInstance);
	}

}
