/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.monitoring.core.jbi;


import java.util.Date;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.spagic.metadb.base.ProcessInstanceStateConstants;
import org.spagic.metadb.dbutils.AuditDBUtils;
import org.spagic.metadb.dbutils.HibernateUtil;
import org.spagic.metadb.model.LogJournal;
import org.spagic.metadb.model.ProcessInstance;
import org.spagic.metadb.model.Property;
import org.spagic.metadb.model.RelevantData;
import org.spagic.metadb.model.Step;
import org.spagic.metadb.model.TransitionInstance;
import org.spagic.monitoring.core.MonitorServiceBase;
import org.spagic.monitoring.message.JBIMessageIFace;
import org.spagic.monitoring.message.MessageIFace;


public class MonitorServiceJBI extends MonitorServiceBase {
	
	private static String IS_STEP_INTERMEDIATE_EVENT = "im.isStepIntermediateEvent";
	//	For a consumer endpoint, indicates if the processor
	//    should store the exchanges in the meta database, for
	//    using the restart feature.
	private boolean storeExchangesForRestart = true;
	
	//    For a consumer endpoint, indicates if the processor
	//    should keep, on the database, only the faulted exchanges
	//    (before the fault).
	//    If false, all ACTIVE exchanges are stored in the database,
	//    not only the faulted exchanges.
	private boolean keepOnlyFaultedExchanges = true;
			
	public MonitorServiceJBI(boolean storeExchangesForRestart,
						  boolean keepOnlyFaultedExchanges, 
						  Log log) {
		super();
		this.storeExchangesForRestart = storeExchangesForRestart;
		this.keepOnlyFaultedExchanges = keepOnlyFaultedExchanges;
		this.log = log;
	}

	/**
	 * Main method: perform the monitoring of all exchanges.
	 * @param exchange
	 */
	public void traceMessage(MessageIFace message) {
		
		Session aSession = null;
		Transaction tx = null;
		try {
			TransitionInstance transitionInstance = null;
			aSession = HibernateUtil.getSessionFactory().openSession();
			tx = aSession.beginTransaction();

			debug("Retrieve the Communication Pattern");
			String communicationPattern = message.getCommunicationPattern();
			debug("Communication Pattern Retrieved ["+communicationPattern+"]");

			// If the state of the exchange is ACTIVE we doesn't search for the
			// exchange on the database because IT CANNOT exist: it is the first exchange
			// of the transition
			if (!message.isFirstMessageForTransition()) {
				debug("Search Transition Instance for a record with ID = " + message.getId());
				transitionInstance = AuditDBUtils.getTransitionInstanceByMessageId(aSession, message.getId());
			}
			
			// Retrieve correlation id to discover the process to which this
			// exchange belongs to
			TransitionManagementState transitionManagementState = null;
			boolean processInstanceCreated = false;	// Setted to true if the instance of the service assembly
												// is created now: useful to limit the iter correlation work
												// only in this case
			
			if (transitionInstance != null) {
				// There is already a Transition with this id logged: we
				// should update the existing row with the state of exchange
				debug(" Found Transition Instance -> Updating... " );
				transitionManagementState = manageExistingTransition(message, transitionInstance);
				debug("State Transition Instance Updated " );
				//and eventually we should save the output message 
				//containing the attachments
				if (!keepOnlyFaultedExchanges) 
					saveTransitionDataOutForAttachments(message, transitionInstance);
			} else {
				// There isn't a Transition with this id logged: we should
				// create a new one	
				debug(" Transition Instance NOT FOUND -> Creating a new one.... " );
				TransitionInstanceCreationResult transitionInstanceCreationResult = createTransitionInstance(message, communicationPattern, aSession);
				transitionInstance = transitionInstanceCreationResult.transitionInstance;
				processInstanceCreated = transitionInstanceCreationResult.processInstanceCreated;
				transitionManagementState = transitionInstanceCreationResult.transitionManagementState;
				debug((processInstanceCreated ? " Created a new Process Instance" : ""));
				
				if (transitionManagementState == null) {
					transitionManagementState = new TransitionManagementState();
				}
				
				if (transitionInstance != null) {
					transitionManagementState.transitionState = transitionInstance.getTransitionState();
				}
			}
			if (transitionInstance != null) {
				aSession.save(transitionInstance);

				// Trace the message and relevant data
				// ------------------------------------
				

				// Retrieve the message to trace
				boolean existsMessageToTrace = message.existsMessageToTrace();

				if (existsMessageToTrace) {
					
					try {
						LogJournal messageInfo = traceMessage(message, transitionInstance, transitionManagementState.transitionState, aSession);
						
						// Trace message relevant data
						List<RelevantData> savedRelevantData = traceRelevantData(messageInfo, aSession);
						
						// Check if the instance of service assembly was just created (and not saved yet)
						// Consider ONLY at least one relevant data have been produced
						if ((processInstanceCreated) && (savedRelevantData.size() > 0)) {
							checkPoolCorrelation(messageInfo, savedRelevantData, aSession);
						}
					} catch (Exception ex) {
			        	log.error(" Exception during tracing of message ", ex);
//			        	log.error(" Exception during tracing of message " + messageToTrace);
					}
					
				}
			}
			
			aSession.flush();
			tx.commit();
			
			// Verify if the process is terminated, only if necessary (only if
			// the current message exchange has status DONE or FAULT)
			if (transitionManagementState.processFault) {
				manageFault(transitionInstance, aSession);
			} else if (transitionManagementState.transitionTerminated) {
				manageDone(transitionInstance, aSession);
			}
			
		} catch (Exception e) {
        	log.error("traceExchange ", e);
			tx.rollback();
		} finally {
			if (aSession != null) {
				aSession.close();
			}
		}
	}
	
	/**
	 * Create on the database a new TransitionInstance that corresponds to a new Message Exchange.
	 * @param message Current Message Exchange
	 * @param communicationPattern Message Exchange Pattern
	 * @param aSession Hibernate Session
	 * @return TransitionInstanceCreationResult result of the creation
	 */
	private TransitionInstanceCreationResult createTransitionInstance(MessageIFace message, String communicationPattern, Session aSession) {
		Step sourceStep = null;
		Step targetStep = null;

		// There isn't a Message Exchange with this id logged: we should
		// create a new one
		debug("createMessageExchange:: START");
		TransitionInstanceCreationResult result = new TransitionInstanceCreationResult();
				
		// Retrieve SOURCE service endpoint: service name (QName) + endpoint name
		QName sourceServiceQName = message.getSourceStepQName();
		if (sourceServiceQName == null) {
			// Error
	    	log.warn("createMessageExchange -> No configured Sender endpoint for exchange:" + message);
	    	// Commented because in some cases we don't have the source service name available
	    	// return result;
		} else {
			String sourceServiceName = sourceServiceQName.getLocalPart();
			sourceServiceName = sourceServiceName.substring(sourceServiceName.indexOf(':') + 1);
			sourceStep = AuditDBUtils.getStepByName(aSession, sourceServiceQName.getNamespaceURI(), sourceServiceName);
			
			if (sourceStep == null) {
				debug("Process not registered on database, sourceStep " + sourceStep + " not found");
				return result;
			} else if (!sourceStep.getProcess().isMonitorEnabled()) {
				debug("Monitoring on process " + sourceStep.getProcess().getName() + " not enabled");
				return result;
			} else {
				Property isIntermediateEventProperty = AuditDBUtils.getStepPropertyByName(aSession, sourceStep, IS_STEP_INTERMEDIATE_EVENT);
				if (isIntermediateEventProperty != null){
					debug("Transition on step " + sourceStep.getProcess().getName() + " is related to event handling");
					return result;
				}
			}
		}

		// Retrieve TARGET service for the exchange
		QName targetServiceName = message.getTargetStepQName();
		targetStep = AuditDBUtils.getStepByName(aSession, targetServiceName.getNamespaceURI(), targetServiceName.getLocalPart());
		debug("createMessageExchange -> Target Step ["+targetStep+"]");

		if (targetStep == null) {
			debug("Process not registered on database, targetStep " + targetStep + " not found");
			debug("targetServiceName.getNamespaceURI(): " + targetServiceName.getNamespaceURI());
			debug("targetServiceName.getLocalPart(): " + targetServiceName.getLocalPart());
			return result;
		} else if (!targetStep.getProcess().isMonitorEnabled()) {
			// The step is replicated (there is the same check for sourceStep) because in some cases
			// the first check is not executed because the sourceServiceName is not available
			debug("Monitoring on process " + targetStep.getProcess().getName() + " not enabled");
			return result;
		}
		
		// Retrieve exchange correlation id to discover the process to
		// which the exchange
		// belongs to
		String correlationID = message.getCorrelationID();
		debug("createMessageExchange:: Correlation ID ["+correlationID+"]");
		ProcessInstance processInstance = AuditDBUtils.getProcessInstanceByIdCorrelation(aSession, correlationID);

		if (processInstance == null) {
			// This is a new process instance: create a new row
			// Before creating a new service assembly instance, I need to retrieve
			// the service assembly, and to retrieve it I can use the endpoint
			processInstance = createProcessInstance(targetStep.getProcess(), correlationID, aSession);
			result.processInstanceCreated = true;
		}

		// The instance of the process already exists:
		// we have to create the journal_instance_service row
		TransitionInstance transitionInstance = new TransitionInstance();
		transitionInstance.setProcessInstance(processInstance);
		transitionInstance.setIdMessage(message.getId());
		
		// Set the exchangeStart retrieving it from the MessageExchange's property
		transitionInstance.setTransitionStart(message.getTransitionStartDate());
		
		// Determine the state for this transition
		result.transitionManagementState = message.getStartState();
		transitionInstance.setTransitionState(result.transitionManagementState.transitionState);
		transitionInstance.setStepByIdStepSource(sourceStep);
		transitionInstance.setStepByIdStepTarget(targetStep);
		
		
		// Store Message Exchange for allow restart in case of errors
		if (storeExchangesForRestart) {
			transitionInstance.setTransitionData(message.getSerializedMessage());
		}
		
		//set the attachment in input message
		int attachmentsIn = ((JBIMessageIFace)message).getAttachmentNames().size();
		transitionInstance.setAttachmentIn(attachmentsIn);
		
		result.transitionInstance = transitionInstance;
		return result;
	}
	
	//should udpate the existing row with the transitiondata containing the attachments
	protected void saveTransitionDataOutForAttachments(MessageIFace message, TransitionInstance transitionInstance){
			((JBIMessageIFace)message).saveMessageOutForAttachments(transitionInstance);
	}
		
	/**
	 * Invoked when a FAULT message is received: it's responsible for updating the process state.
	 * @param transitionInstance Current TransitionInstance
	 * @param aSession Hibernate session
	 */
	private void manageFault(TransitionInstance transitionInstance, Session aSession) {
		Transaction finalTx = null;

		try {
			ProcessInstance processInstance = transitionInstance.getProcessInstance();
			
			// Check if the process state is ACTIVE yet
			if (processInstance.getState() == ProcessInstanceStateConstants.PROCESS_ACTIVE) {
				// Begin a new transaction to close the process state
				finalTx = aSession.beginTransaction();

				// The process has faulted: update its status
				processInstance.setState(ProcessInstanceStateConstants.PROCESS_FAULTED);
				
				Date endDate = new Date();
				Date startDate = processInstance.getStartdate();
				processInstance.setEnddate(endDate);
				processInstance.setExecutionTime(endDate.getTime() - startDate.getTime());
				
				/*Long idServiceAssemblySaved = (Long) */ aSession.save(processInstance);
				aSession.flush();
				finalTx.commit();
			}
		} catch (Exception ex) {
        	log.error("manageFault ", ex);
        	log.error("manageFault, TransitionInstance = " + transitionInstance);
			finalTx.rollback();
		}
	}
	
	/**
	 * Invoked when a DONE message is received: it's responsible for updating the process state.
	 * @param transitionInstance Current TransitionInstance
	 * @param aSession Hibernate session
	 */
	private void manageDone(TransitionInstance transitionInstance, Session aSession) {
		Transaction finalTx = null;

		try {
			ProcessInstance processInstance = transitionInstance.getProcessInstance();
			
			// Begin a new transaction to close the process state
			finalTx = aSession.beginTransaction();

			// If the exchange is done, we can remove it from database because
			// we dont need anymore the restart feature, for this exchange
			if (keepOnlyFaultedExchanges) {
				transitionInstance.setTransitionData(null);
				aSession.save(transitionInstance);
			}
			
			// Check if the process state is ACTIVE yet
			if (processInstance.getState() == ProcessInstanceStateConstants.PROCESS_ACTIVE) {

				long counter = AuditDBUtils.getTransitionInstancesNotClosed(aSession, processInstance);
				boolean isWaiting = AuditDBUtils.isWaiting(aSession, processInstance);

				if ((counter == 0) && (!isWaiting)){
					// The process should be finished: update its status
					processInstance.setState(ProcessInstanceStateConstants.PROCESS_DONE);
					Date endDate = new Date();
					Date startDate = processInstance.getStartdate();
					processInstance.setEnddate(endDate);
					processInstance.setExecutionTime(endDate.getTime() - startDate.getTime());
					
				}
			}
			
			aSession.flush();
			finalTx.commit();
		} catch (Exception ex) {
        	log.error("manageDone ", ex);
        	log.error("manageDone, TransitionInstance = " + transitionInstance);
			finalTx.rollback();
		}
	}
	
	public void debug(Object o){
		if (log.isDebugEnabled()){
			log.debug(o);
		}
	}

}
