/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.monitoring.ode.listeners;


import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ode.bpel.evt.ActivityEvent;
import org.apache.ode.bpel.evt.ActivityExecEndEvent;
import org.apache.ode.bpel.evt.ActivityExecStartEvent;
import org.apache.ode.bpel.evt.ActivityFailureEvent;
import org.apache.ode.bpel.evt.BpelEvent;
import org.apache.ode.bpel.evt.EventContext;
import org.apache.ode.bpel.evt.ProcessCompletionEvent;
import org.apache.ode.bpel.evt.ScopeCompletionEvent;
import org.apache.ode.bpel.evt.ScopeEvent;
import org.apache.ode.bpel.evt.ScopeStartEvent;
import org.apache.ode.bpel.evt.VariableModificationEvent;
import org.apache.ode.bpel.iapi.BpelEventListener;
import org.spagic.monitoring.core.bpel.MonitorServiceBPEL;
import org.spagic.monitoring.ode.message.OdeMessage;

/**
 * @author boccalon
 * Monitoring listener specific for ODE BPEL engine, to be used in Spagic
 *
 */
public class OdeListener implements BpelEventListener {

	private static final Log log = LogFactory.getLog(OdeListener.class);
	
	public static final String RECEIVE_ACTIVITY = "OPickReceive";
	public static final String ASSIGN_ACTIVITY = "OAssign";
	public static final String INVOKE_ACTIVITY = "OInvoke";
	public static final String FLOW_ACTIVITY = "OFlow";
	public static final String REPLY_ACTIVITY = "OReply";
	
	private Map<Long, Map<Long, Map<String, String>>> variablesData = null;  // Map of ProcessID(long), [Map of ScopeID(long), {Map of variable name, variable value}]
	private Map<Long, List<EventContext>> contextsMap = null; // Map of ProcessID(long), [List of EventContext]
	private long startupTime = 0;	// Listener startup time
	private MonitorServiceBPEL monitorService = null; // Monitor engine	

	public OdeListener() {
		super();
		
		log.debug("Build Ode listener");
		
		// Map for storing intermediate variables values
		variablesData = new HashMap<Long, Map<Long, Map<String, String>>>();
		
		// Map for storing all scopes
		contextsMap = new HashMap<Long, List<EventContext>>();
		
		// Listener startup date
		startupTime = (new Date()).getTime();
		
		// Instantiate monitor service only once (should be thread safe)
		monitorService = new MonitorServiceBPEL(false/*storeExchangesForRestart*/, log);
	}
	
	/**
	 * Not all events are interesting for the listener: this method checks which ones
	 * are relevant.
	 * @param ae The event
	 * @return True if the event has to be traced
	 */
	public boolean isActivityEventInteresting(ActivityEvent ae) {
		String activityType = ae.getActivityType();
		return (activityType.equals(RECEIVE_ACTIVITY) ||
				activityType.equals(ASSIGN_ACTIVITY) ||
				activityType.equals(INVOKE_ACTIVITY) ||
				activityType.equals(FLOW_ACTIVITY) ||
				activityType.equals(REPLY_ACTIVITY));
	}

	/* (non-Javadoc)
	 * @see org.apache.ode.bpel.iapi.BpelEventListener#onEvent(org.apache.ode.bpel.evt.BpelEvent)
	 */
	public void onEvent(BpelEvent event) {
		try {
			// Current process id
			Long pId = null;
			// Current scope id
			Long scopeId = null;
			
			log.debug("Tracing event:");
			log.debug(event);
			
			// Manage only interesting events
			boolean traceEvent = false;
			if (event instanceof ProcessCompletionEvent) {
				// Remove process variables, if any, at the end of tracing activity
				pId = ((ProcessCompletionEvent)event).getProcessInstanceId();
				
				// Allow process state update on database
				traceEvent = true;
			} else if ((event instanceof ActivityExecStartEvent) ||
					   (event instanceof ActivityExecEndEvent) ||
					   (event instanceof ActivityFailureEvent)) {
				
				ActivityEvent ae = (ActivityEvent)event;
				if (isActivityEventInteresting(ae)) {
					pId = ae.getProcessInstanceId();
					scopeId = ae.getScopeId();
					traceEvent = true;
				}
				
			} else if (event instanceof VariableModificationEvent) {
				Map<Long, Map<String, String>> currentProcessVariablesMap = null;
				Map<String, String> currentScopeMap = null;
				VariableModificationEvent vme = (VariableModificationEvent) event;

				// Retrieve process id
				pId = vme.getProcessInstanceId();

				// Get the value of the modified variable
				String varName = vme.getVarName();
				String varValue = vme.eventContext.getVariableData(varName);

				log.debug("Variable changed. Variable name: " + varName);
				
				// Check if we were able to get the variable value
				if (varValue == null) {
					// The variable is defined in a scope different from the current one:
					// search in all available contexts
					// Retrieve all contexts for current process
					List<EventContext> processContexts = contextsMap.get(pId);
					log.debug("Found " + processContexts.size() + " contexts for process: " + pId);
					for (EventContext context : processContexts) {
						log.debug("Searching variable: " + varName + " in context: " + context.getScopeInstanceId());
						varValue = context.getVariableData(varName);
						if (varValue != null) {
							// Found !!!
							log.debug("Found variable in context with scope instance id: " + context.getScopeInstanceId());
							break;
						} else {
							log.debug("Variable: " + varName + " not found in context: " + context.getScopeInstanceId());
						}
					}
				}
				if (varValue == null) {
					log.error("Unable to retrieve variable value for event: " + vme);
				} else {
					log.debug("Variable value: " + varValue);

					// Save variable value for later use
					
					// Retrieve variables for this process
					currentProcessVariablesMap = variablesData.get(pId);
					if (currentProcessVariablesMap == null) {
						currentProcessVariablesMap = new HashMap<Long, Map<String, String>>();
						variablesData.put(pId, currentProcessVariablesMap);
						log.debug("Created variable map for process: " + pId);
					} else {
						log.debug("Retrieved variable map for process: " + pId);
					}
					
					// Retrieve variables for this scope
					currentScopeMap = currentProcessVariablesMap.get(vme.getScopeId());
					if (currentScopeMap == null) {
						currentScopeMap = new HashMap<String, String>();
						currentProcessVariablesMap.put(vme.getScopeId(), currentScopeMap);
						log.debug("Created scope map for scope id " + vme.getScopeId() + " and process: " + pId);
					} else {
						log.debug("Retrieved scope map for scope id " + vme.getScopeId() + " and process: " + pId);
					}
					
					// Stores variable in the current scope
					currentScopeMap.put(varName, varValue);
				}
			} else if (event instanceof ScopeStartEvent) {
				ScopeStartEvent sse = (ScopeStartEvent)event;
				
				// Retrieve process id
				pId = sse.getProcessInstanceId();
				
				log.debug("Scope created for process id: " + pId + ", scope id: " + sse.getScopeId());

				// Retrieve variables for this process
				List<EventContext> processContextList = contextsMap.get(pId);
				if (processContextList == null) {
					processContextList = new ArrayList<EventContext>();
					contextsMap.put(pId, processContextList);
				}
				
				// Store context in first position, so the search begins from this one
				processContextList.add(0, sse.eventContext);
			} else if (event instanceof ScopeCompletionEvent) {
				ScopeCompletionEvent sse = (ScopeCompletionEvent)event;
				List<EventContext> processContextList = contextsMap.get(pId);
				boolean deleted = processContextList.remove(sse.eventContext);

				log.debug("Scope terminated for process id: " + pId + ", scope id: " + sse.getScopeId() + ", deleted: " + deleted);
			}
			
			if (traceEvent) {
				log.debug("Tracing event..");
				
				// Map of all scopes of the current process (there are different scopes
				// within a process)
				Map<Long, Map<String, String>> currentProcessVariablesMap = null;
				// Map of all variables within current scope
				Map<String, String> currentScopeMap = null;

				// Retrieve variables for this process
				currentProcessVariablesMap = variablesData.get(pId);
				if (currentProcessVariablesMap != null) {
					
					// Retrieve variables for this scope
					currentScopeMap = getFirstNotNullScope(scopeId, currentProcessVariablesMap, event);
					
					log.debug("Retrieved variables map for process: " + pId);
					
					try {
						OdeMessage bpelMessage = new OdeMessage(event, log, currentScopeMap, startupTime);
				    	monitorService.traceMessage(bpelMessage);
						log.debug("Event traced");
					} finally {
						if (event instanceof ProcessCompletionEvent) {
							// Remove process variables, if any
							pId = ((ProcessCompletionEvent)event).getProcessInstanceId();
							variablesData.remove(pId);
							contextsMap.remove(pId);
												
							log.debug("Process " + pId + " completed");
						}
					}
				} else {
					log.debug("Unable to retrieve variables map for process: " + pId);
				}
				
			}
		} catch (Throwable ex) {
			// An exception should NEVER thrown by this code, otherwise the process is
			// not executed
			log.error("Problems during process execution", ex);
		}
	}
	
	/**
	 * @param scopeId Start scope id for the search
	 * @param currentProcessVariablesMap Map of the scopes for the current process
	 * @return The first not null scope retrieved navigating through the chain of parent scopes
	 */
	private Map<String, String> getFirstNotNullScope(Long scopeId, Map<Long, Map<String, String>> currentProcessVariablesMap, BpelEvent event) {
		Map<String, String> currentScopeMap = currentProcessVariablesMap.get(scopeId);
		if (currentScopeMap == null) {
			if (event instanceof ScopeEvent) {
				// We don't store the hierarchy of the scopes, so the only choice that
				// we have is trying with the parent scope. Storing the hierarchy of scope
				// should be a better choice in future.
				scopeId = ((ScopeEvent)event).getParentScopeId();
				currentScopeMap = currentProcessVariablesMap.get(scopeId);
			}
		}
		return currentScopeMap;
	}

	/* (non-Javadoc)
	 * @see org.apache.ode.bpel.iapi.BpelEventListener#shutdown()
	 */
	public void shutdown() {
		variablesData = null;
		contextsMap = null;
		log.debug("Shutdown Ode listener");
	}

	/* (non-Javadoc)
	 * @see org.apache.ode.bpel.iapi.BpelEventListener#startup(java.util.Properties)
	 */
	public void startup(Properties props) {
		log.debug("Startup Ode listener");
	}

}
