/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.monitoring.ode.message;


import java.util.Date;
import java.util.Map;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.ode.bpel.evt.ActivityEvent;
import org.apache.ode.bpel.evt.ActivityExecEndEvent;
import org.apache.ode.bpel.evt.ActivityExecStartEvent;
import org.apache.ode.bpel.evt.ActivityFailureEvent;
import org.apache.ode.bpel.evt.BpelEvent;
import org.apache.ode.bpel.evt.ProcessCompletionEvent;
import org.apache.ode.bpel.evt.ProcessInstanceEvent;
import org.spagic.metadb.base.TransitionStateConstants;
import org.spagic.metadb.model.TransitionInstance;
import org.spagic.metadb.model.TransitionState;
import org.spagic.monitoring.core.TransitionStateCache;
import org.spagic.monitoring.core.MonitorIService.TransitionManagementState;
import org.spagic.monitoring.message.BPELMessage;

public class OdeMessage implements BPELMessage {

	public static final String ODE_PREFIX = "ODE:";

	private BpelEvent event;
	private Log log;
	private Map<String, String> variablesData;
	private long listenerStartupTime;
	
	public OdeMessage(BpelEvent event, Log log, Map<String, String> variablesData, long listenerStartupTime) {
		this.event = event;
		this.log = log;
		this.variablesData = variablesData;
		this.listenerStartupTime = listenerStartupTime;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#existsMessageToTrace()
	 */
	public boolean existsMessageToTrace() {
		return (((event instanceof ActivityExecStartEvent) ||
				 (event instanceof ActivityFailureEvent)) && 
				(variablesData != null));
//		return (((event instanceof ActivityExecEndEvent) ||
//				 (event instanceof ActivityFailureEvent)) && 
//				(variablesData != null));
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getCommunicationPattern()
	 */
	public String getCommunicationPattern() {
		// Not applicable to BPEL
		return null;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getCorrelationID()
	 */
	public String getCorrelationID() {
		// Add a ODE_PREFIX to the correlation id to avoid possible conflicts
		// between different engines that can use the same ids
		
		// This id should be unique for each process: the problem is that if the
		// process in running "in-memory" at each restart of ServiceMix, the
		// process id restart from 0.
		// To avoid the problem I add also the listener startup time (in milliseconds)
		
		if (event instanceof ProcessInstanceEvent) {
			return (ODE_PREFIX + ((ProcessInstanceEvent)event).getProcessInstanceId() + ":" + listenerStartupTime);
		} else {
			log.info("Correlation id not retrievable for event: ");
			log.info(event);
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getTransitionStartDate()
	 */
	public Date getTransitionStartDate() {
		return event.getTimestamp();
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getTransitionCurrentStepDate()
	 */
	public Date getTransitionCurrentStepDate() {
		return event.getTimestamp();
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getId()
	 */
	public String getId() {
		if (event instanceof ActivityEvent) {
			ActivityEvent ae = (ActivityEvent)event;
			// The process instance id and activity id are NOT UNIQUE for processes
			// running in-memory, so we add a timestamp.
			// The timestamp is the listener startup time (in milliseconds)
			return (ODE_PREFIX + ae.getProcessInstanceId() + ":" + ae.getActivityId() + ":" + listenerStartupTime);
		} else {
			log.error("Message id not retrievable for event: ");
			log.error(event);
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getRawMessage()
	 */
	public Object getRawMessage() {
		return event;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getSerializedMessage()
	 */
	public byte[] getSerializedMessage() {
		// Not applicable to BPEL: restart not possible on BPEL
		return null;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getSourceStepQName()
	 */
	public QName getSourceStepQName() {
		// Not retrievable in BPEL
		return null;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getStartState()
	 */
	public TransitionManagementState getStartState() {
		TransitionManagementState startState = new TransitionManagementState();
		
		int transitionState = -1;
		if (event instanceof ActivityExecStartEvent) {
			transitionState = TransitionStateConstants.BPEL_START_ACTIVITY;
		} else if (event instanceof ActivityExecEndEvent) {
			transitionState = TransitionStateConstants.BPEL_END_ACTIVITY;
			startState.transitionTerminated = true;
		} else if (event instanceof ActivityFailureEvent) {
			transitionState = TransitionStateConstants.BPEL_FAULTED_ACTIVITY;
			startState.processFault = true;
		} else {
			log.error("Unable to retrieve start state for event: " + event);
		}
		startState.transitionState = TransitionStateCache.getTransitionState(transitionState);
		return startState;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getTargetStepQName()
	 */
	public QName getTargetStepQName() {
		if (event instanceof ActivityEvent) {
			ActivityEvent ae = (ActivityEvent)event;
			
			QName processName = ae.getProcessName();
			String activityName = ae.getActivityName();
			
			return new QName(processName.getNamespaceURI(), activityName);
		} else {
			log.error("TargetStepQName not retrievable for event: " + event);
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#getXMLMessage()
	 */
	public String getXMLMessage() throws Exception {
		StringBuffer xmlBuffer = new StringBuffer();
		if (event instanceof ActivityFailureEvent) {
			xmlBuffer.append("<ERROR message=\"");
			xmlBuffer.append(((ActivityFailureEvent)event).getFailureReason());
			xmlBuffer.append("\"/>");			
		} else {
			// In the variables map we may have more than one variable
			// We create a XML containing all variable values, with a predefined
			// schema
			xmlBuffer.append("<VARIABLES>");
			for (Map.Entry<String, String> entry : variablesData.entrySet()) {
				xmlBuffer.append("<VARIABLE name=\""+entry.getKey()+"\">");
				xmlBuffer.append("<VALUE>");
				// Remove the XML header
				xmlBuffer.append(entry.getValue().replaceFirst("<\\?xml .*>(\\r\\n)*", ""));
				xmlBuffer.append("</VALUE>");
				xmlBuffer.append("</VARIABLE>\\r\\n");
			}
			xmlBuffer.append("</VARIABLES>");
		}
		return xmlBuffer.toString();
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#isFirstMessageForTransition()
	 */
	public boolean isFirstMessageForTransition() {
		if ((event instanceof ActivityExecStartEvent) ||
			(event instanceof ProcessCompletionEvent))	{
			return true;
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.core.MessageIFace#nextState(org.spagic.metadb.model.TransitionInstance)
	 */
	public TransitionManagementState nextState(TransitionInstance transitionInstance) {
		
		TransitionManagementState newTransitionState = new TransitionManagementState();
		Date sentMessage = getTransitionCurrentStepDate();

		// The state can be only Done
		if (event instanceof ActivityExecEndEvent) {
			// Activity terminated
			TransitionState transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.BPEL_END_ACTIVITY);
			transitionInstance.setTransitionState(transitionState);
			// Mark the exchange as done, to check if the process is terminated
			newTransitionState.transitionTerminated = true;
		} else if (event instanceof ActivityFailureEvent) {
			// Activity faulted
			TransitionState transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.BPEL_FAULTED_ACTIVITY);
			transitionInstance.setTransitionState(transitionState);
			
			// Mark the transition as faulted, to check if the process is terminated
			newTransitionState.processFault = true;
		} else {
			log.error("Unable to determine nextState for event: ");
			log.error(event);
		}
		
		transitionInstance.setTransitionEnd(sentMessage);
		newTransitionState.transitionState = transitionInstance.getTransitionState();
		return newTransitionState;
	}

	/* (non-Javadoc)
	 * @see org.spagic.monitoring.message.BPELMessage#isProcessTerminated()
	 */
	public boolean isProcessTerminated() {
		return (event instanceof ProcessCompletionEvent);
	}

}
