package org.spagic.monitoring.smx.message;


import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Calendar;
import java.util.Date;
import java.util.Set;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jms.Message;
import javax.xml.namespace.QName;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.commons.logging.Log;
import org.apache.servicemix.JbiConstants;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.spagic.metadb.base.TransitionStateConstants;
import org.spagic.metadb.model.TransitionInstance;
import org.spagic.metadb.model.TransitionState;
import org.spagic.monitoring.core.CommunicationPatterns;
import org.spagic.monitoring.core.TransitionStateCache;
import org.spagic.monitoring.core.MonitorIService.TransitionManagementState;
import org.spagic.monitoring.message.JBIMessageIFace;
import org.spagic.monitoring.message.MessageIFace;
import org.xml.sax.SAXException;


public class JBIMessage implements JBIMessageIFace {

	public static final String MEP_INONLY = "INONLY";
	public static final String MEP_INOUT = "INOUT";
	public static final String MEP_ROBUSTINONLY = "ROBUSTINONLY";
	public static final String MEP_INOPTIONALOUT = "INOPTIONALOUT";

	public static final String SERVICEMIX_MEP_INONLY = "org.apache.servicemix.jbi.messaging.InOnlyImpl";
	public static final String SERVICEMIX_MEP_INOUT = "org.apache.servicemix.jbi.messaging.InOutImpl";
	public static final String SERVICEMIX_MEP_ROBUSTINONLY = "org.apache.servicemix.jbi.messaging.RobustInOnlyImpl";
	public static final String SERVICEMIX_MEP_INOPTIONALOUT = "org.apache.servicemix.jbi.messaging.InOptionalOutImpl";

	public static final String SMX_PREFIX = "SMX:";
	
	private MessageExchange exchange;
	private NormalizedMessage message;
	private Log log;
	private TransitionManagementState transitionManagementState;
	private String communicationPattern;
	
	public JBIMessage(MessageExchange exchange, Log log) {
		super();
		this.exchange = exchange;
		this.log = log;
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getId()
	 */
	public String getId() {
		return SMX_PREFIX + exchange.getExchangeId();
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getRawMessage()
	 */
	public Object getRawMessage() {
		return exchange;
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getCommunicationPattern()
	 */
	public String getCommunicationPattern() {
		if (communicationPattern != null) {
			return communicationPattern;
		}
		
		String mepClassName = exchange.getClass().getName();
		if (mepClassName.equals(SERVICEMIX_MEP_INONLY)) {
			communicationPattern = CommunicationPatterns.JBI_INONLY;
		} else if (mepClassName.equals(SERVICEMIX_MEP_INOUT)) {
			communicationPattern = CommunicationPatterns.JBI_INOUT;
		} else if (mepClassName.equals(SERVICEMIX_MEP_ROBUSTINONLY)) {
			communicationPattern = CommunicationPatterns.JBI_ROBUSTINONLY;
		} else if (mepClassName.equals(SERVICEMIX_MEP_INOPTIONALOUT)) {
			communicationPattern = CommunicationPatterns.JBI_INOPTIONALOUT;
		} else {
			// Errore
			log.error("MEP not supported: " + communicationPattern);
		}
		return communicationPattern;
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getSourceStepQName()
	 */
	public QName getSourceStepQName() {
		QName sourceStepQName = null;
		String sourceServiceEndpointName = (String)exchange.getProperty(JbiConstants.SENDER_ENDPOINT);
		debug("Sender Endpoint ["+sourceServiceEndpointName+"]");
		if (sourceServiceEndpointName != null) {
			try {
				sourceStepQName = QName.valueOf(sourceServiceEndpointName);
			} catch (IllegalArgumentException ex) {
				log.error("Cannot retrieve QName from step name: " + sourceServiceEndpointName, ex);
			}
		}
		return sourceStepQName;
	}
	
	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getTargetStepQName()
	 */
	public QName getTargetStepQName() {
		QName targetServiceName = exchange.getService();
		return targetServiceName;
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getCorrelationID()
	 */
	public String getCorrelationID() {
		// Add a CORRELATION_ID_PREFIX to the correlation id to avoid possible conflicts
		// between different engines that can use the same ids
		String correlationID = SMX_PREFIX + (String) exchange.getProperty(JbiConstants.CORRELATION_ID);
		return correlationID;
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getDate()
	 */
	public Date getTransitionStartDate() {
		Date result = null;
		Object date =  getNormalizedMessage().getProperty(JbiConstants.DATESTAMP_PROPERTY_NAME);
		if (date instanceof Calendar) 
			result = ((Calendar)date).getTime();
		else if (date instanceof Date) 
			result = (Date) date;
		return result;
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getForwardingDate()
	 */
	public Date getTransitionCurrentStepDate() {
		NormalizedMessage nmrMessage = getNormalizedMessage();
		Date sentMessage = (nmrMessage != null) ? (Date) nmrMessage
				.getProperty(JbiConstants.DATESTAMP_PROPERTY_NAME) : new Date();
		return sentMessage;
	}
	
	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getState()
	 */
	public TransitionManagementState getStartState() {
		if (transitionManagementState != null) {
			return transitionManagementState;
		}
		transitionManagementState = new TransitionManagementState();
		
		String communicationPattern = getCommunicationPattern();
		if (communicationPattern.equals(CommunicationPatterns.JBI_INONLY)) {
			transitionManagementState.transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INONLY_ACTIVE);
		} else if (communicationPattern.equals(CommunicationPatterns.JBI_ROBUSTINONLY)) {
			transitionManagementState.transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_ROBUSTINONLY_ACTIVE);
		} else if (communicationPattern.equals(CommunicationPatterns.JBI_INOUT)) {
			transitionManagementState.transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INOUT_ACTIVE);
		} else if (communicationPattern.equals(CommunicationPatterns.JBI_INOPTIONALOUT)) {
			transitionManagementState.transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INOPTIONALOUT_ACTIVE);
		}

		return transitionManagementState;
	}

	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#getSerializedMessage()
	 */
	public byte[] getSerializedMessage() {
		return serializeExchange(exchange);
	}

	public String getXMLMessage() throws IOException, SAXException, MessagingException, TransformerException, ParserConfigurationException {
		SourceTransformer sourceTransformer = new SourceTransformer();
		NormalizedMessage nmrMessage = getNormalizedMessage();
		String xmlMessage = sourceTransformer.contentToString(nmrMessage);

		// Recreate the content again, if there was a content
		nmrMessage.setContent((xmlMessage != null) ? new StringSource(xmlMessage)
				: null);
		// Dont never return a null object
		return (xmlMessage != null) ? xmlMessage : "";
	}
	
	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#existsMessageToTrace()
	 */
	public boolean existsMessageToTrace() {
		return (getNormalizedMessage() != null);
	}
		
	/**
	 * Extracts ,from the current exchange, the message to trace on database: it can
	 * be the in message, the out message, the done message, etc
	 * @param exchange Current exchange
	 * @param mep Message Exchange Pattern
	 * @return NormalizedMessage to trace
	 */
	private NormalizedMessage getNormalizedMessage() {
		if (message != null) {
			return message;
		}
		String communicationPattern = getCommunicationPattern();
		if (communicationPattern.equals(CommunicationPatterns.JBI_INONLY)) {
			// The state can be only Active or Done
			debug(" MEP is IN_ONLY ");
			if (exchange.getStatus().equals(ExchangeStatus.ACTIVE) ||
				exchange.getStatus().equals(ExchangeStatus.DONE)) {
				NormalizedMessage inMessage = ((InOnly) exchange).getInMessage();
				message = inMessage;
			} else {
				// Error
				Fault faultDetail = ((InOnly) exchange).getFault();
				Exception ex = ((InOnly) exchange).getError();
	        	debug("retrieveMessageToTrace -> Fault = " + faultDetail);
	        	log.debug("retrieveMessageToTrace ->", ex);
	        	
	        	if (faultDetail != null) {
					message = faultDetail;
				} else if (ex != null) {
					// Create a dummy message to allow storing the error
					try {
						NormalizedMessage errorMessage = exchange.createMessage();
						errorMessage.setContent(new StringSource((ex.getMessage() != null) ? ex.getMessage() : ex.toString()));
						errorMessage.setProperty(JbiConstants.DATESTAMP_PROPERTY_NAME, new Date());
						message = errorMessage;
					} catch (MessagingException mex) {
						log.error("Error creating error message", mex);
					}
				} else {
					log.error("No error to trace: both getFault and getError return null");
				}
			}
		} else if (communicationPattern.equals(CommunicationPatterns.JBI_INOUT)) {
			debug(" MEP is IN_OUT ");
			if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
				
				// Check if there are errors: possible also if state is ACTIVE
				Fault faultDetail = ((InOut) exchange).getFault();
				if (faultDetail != null) {
					debug(" Message is FAULT");
					message = faultDetail;
				}
				// I must check if this is the request or the response
				else if (exchange.getRole() == MessageExchange.Role.CONSUMER) {
					debug(" Role is CONSUMER ");
					NormalizedMessage outMessage = ((InOut) exchange).getInMessage();
					message = outMessage;
				} else if (exchange.getRole() == MessageExchange.Role.PROVIDER) {
					debug(" Role is PROVIDER ");
					NormalizedMessage inMessage = ((InOut) exchange).getOutMessage();
					message = inMessage;
				} else {
					// Error
		        	log.error("retrieveMessageToTrace. Mep type=MEP_INOUT, role unknown");
				}
			} else if (exchange.getStatus().equals(ExchangeStatus.DONE)) {
				NormalizedMessage inMessage = ((InOut) exchange).getInMessage();
				message = inMessage;
			} else if (exchange.getStatus().equals(ExchangeStatus.ERROR)) {
				// Error
				Fault faultDetail = ((InOut) exchange).getFault();
				Exception ex = ((InOut) exchange).getError();
	        	debug("retrieveMessageToTrace -> Fault = " + faultDetail);
	        	log.debug("retrieveMessageToTrace ->", ex);
	        	
	        	if (faultDetail != null) {
					message = faultDetail;
				} else if (ex != null) {
					// Create a dummy message to allow storing the error
					try {
						StringBuffer sb = new StringBuffer();
						sb.append("<MESSAGE SENT=\"FALSE\" ERROR=\"");
						sb.append(ex.getMessage());
						sb.append("\"/>");
						
						Fault fault = exchange.createFault();
						fault.setContent(new StringSource(sb.toString()));
						fault.setProperty(JbiConstants.DATESTAMP_PROPERTY_NAME, new Date());
						message = fault;
					} catch (MessagingException mex) {
						log.error("Error creating error message", mex);
					}
				} else {
					log.error("No error to trace: both getFault and getError return null");
				}
			} else {
				// Error
	        	log.error("retrieveMessageToTrace. Mep type=MEP_INOUT, status undefined ");
			}
		} else if (communicationPattern.equals(CommunicationPatterns.JBI_ROBUSTINONLY)) {
			debug(" MEP is ROBUST_IN_ONLY ");
		} else if (communicationPattern.equals(CommunicationPatterns.JBI_INOPTIONALOUT)) {
			debug(" MEP is ROBUST IN_OPTIONAL_OUT ");
		} else {
			// Errore
        	log.error("retrieveMessageToTrace. Mep type=UNDEFINED ");
		}
		return message;
	}

	/**
	 * Serializes an exchange on a byte array
	 * @param exchange Exchange to serialize
	 * @return Serialized exchange
	 */
	private byte[] serializeExchange(MessageExchange exchange) {
        byte[] buf = null;
        
        // Serialize to a byte array
		try {
	        ByteArrayOutputStream bos = new ByteArrayOutputStream();
	        ObjectOutputStream out = new ObjectOutputStream(bos);
	        out.writeObject(exchange);
	        out.close();
	    
	        // Get the bytes of the serialized object
	        buf = bos.toByteArray();
		} catch (Exception ex) {
			log.error("Error serializing exchange " + exchange, ex );
		}
        return buf;
	}
	
	private byte[] serializeNormalizedMessage() {
        byte[] buf = null;
        
        // Serialize to a byte array
		try {
	        ByteArrayOutputStream bos = new ByteArrayOutputStream();
	        ObjectOutputStream out = new ObjectOutputStream(bos);
	        out.writeObject(this.getNormalizedMessage());
	        out.close();
	    
	        // Get the bytes of the serialized object
	        buf = bos.toByteArray();
		} catch (Exception ex) {
			log.error("Error serializing normalized message " + exchange, ex );
		}
        return buf;
	}
	
	/**
	 * Deserialize a byte stream to rebuild a Message Exchange
	 * @param buf Byte stream
	 * @return MessageExchange
	 */
	private MessageExchange deserializeExchange(byte[] buf) {
		MessageExchange messageExchange = null;
		if (buf == null) 
			return null;
		try {
			// Deserialize from a byte array
			ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(buf));
			messageExchange = (MessageExchange)in.readObject();
	        in.close();
		} catch (Exception ex) {
			log.error("Error serializing exchange " + messageExchange, ex );
		}
        return messageExchange;
	}
	
	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#nextState(org.spagic.metadb.model.TransitionInstance)
	 * NOTE:
	 * Developing with servicemix we encounter these possibility to notify error by components:
	 * 1 - maintain the Active status and provide the fault message
	 * 2 - set the Error status and provide an error message (+ fault message?)
	 * 
	 * Is not clear if point 1 is JBI compliant.

	 */
	public TransitionManagementState nextState(TransitionInstance transitionInstance) {
		TransitionManagementState newTransitionState = null;
		String communicationPattern = getCommunicationPattern();

		if (communicationPattern.equals(MEP_INONLY)) {
			newTransitionState = nextStateForInOnly(transitionInstance);
		} else if (communicationPattern.equals(MEP_INOUT)) {
			newTransitionState = nextStateForInOut(transitionInstance);
		} else if (communicationPattern.equals(MEP_ROBUSTINONLY)) {
			newTransitionState = nextStateForRobustInOnly(transitionInstance);
		} else if (communicationPattern.equals(MEP_INOPTIONALOUT)) {
			newTransitionState = nextStateForInOptionalOut(transitionInstance);
		} else {
			// Errore
			log.error("MEP not supported: " + communicationPattern);
		}
		return newTransitionState;
	}

	private TransitionManagementState nextStateForInOnly(TransitionInstance transitionInstance) {
		
		TransitionManagementState newTransitionState = new TransitionManagementState();
		MessageExchange exchange = (MessageExchange)getRawMessage();
		Date sentMessage = getTransitionCurrentStepDate();

		// The state can be only Done
		if (exchange.getStatus().equals(ExchangeStatus.DONE) && (exchange.getError() == null)) {
			// Exchange terminated
			TransitionState transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INONLY_DONE);
			transitionInstance.setTransitionState(transitionState);
			// Mark the exchange as done, to check if the process is terminated
			newTransitionState.transitionTerminated = true;
		} else {
			// Exchange terminated with ERROR: this MEP state is not a standard JBI state
			// We introduced this state to manage properly the error case with InOnly MEP
			TransitionState transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INONLY_FAULT);
			transitionInstance.setTransitionState(transitionState);
			
			// Mark the exchange as faulted, to check if the process is terminated
			newTransitionState.processFault = true;
		}
		transitionInstance.setTransitionEnd(sentMessage);
		newTransitionState.transitionState = transitionInstance.getTransitionState();
		return newTransitionState;
	}
	
	private TransitionManagementState nextStateForInOut(TransitionInstance transitionInstance) {
		
		TransitionManagementState newTransitionState = new TransitionManagementState();
		MessageExchange exchange = (MessageExchange)getRawMessage();
		Date sentMessage = getTransitionCurrentStepDate();
		NormalizedMessage faultMessage = ((InOut) exchange).getFault();

		if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
			TransitionState transitionState = null;
			
			if (faultMessage == null) {
				NormalizedMessage outMessage = ((InOut) exchange).getOutMessage();
				if (outMessage == null) {
					// Errore
					log.error("Both Out and Fault messages empty");
				}
				transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INOUT_OUT);
			} else {
				// Fault
				transitionInstance.setTransitionEnd(sentMessage);
				transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INOUT_FAULT);
			}
			transitionInstance.setTransitionState(transitionState);
			newTransitionState.transitionState = transitionInstance.getTransitionState();

		} else if (exchange.getStatus().equals(ExchangeStatus.DONE)) {

			// If there are errors, we dont mark the process as DONE in the database, so it's visualized
			// with an error state on the console
			if (faultMessage == null) {
				TransitionState transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INOUT_DONE);
				transitionInstance.setTransitionState(transitionState);
			} else {
				// Keep previous state: should be FAULT
			}
			
			// Real exchange status is DONE
			newTransitionState.transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INOUT_DONE);
			
			// Mark the exchange as done, to check if the process is terminated
			newTransitionState.transitionTerminated = true;
			transitionInstance.setTransitionEnd(sentMessage);
		} else {
			// Error
			TransitionState transitionState = TransitionStateCache.getTransitionState(TransitionStateConstants.JBI_MEP_INOUT_FAULT);
			transitionInstance.setTransitionState(transitionState);
			transitionInstance.setTransitionEnd(sentMessage);
			newTransitionState.transitionState = transitionInstance.getTransitionState();
			newTransitionState.processFault = true;
		}
		
		return newTransitionState;
	}
	
	private TransitionManagementState nextStateForRobustInOnly(TransitionInstance transitionInstance) {
		return null;
	}
	
	private TransitionManagementState nextStateForInOptionalOut(TransitionInstance transitionInstance) {
		return null;
	}	
	
	/* (non-Javadoc)
	 * @see it.eng.spagosoa.monitoring.MessageIFace#isFirstMessageForTransition()
	 */
	public boolean isFirstMessageForTransition() {
		return firstExchangeForTransition((MessageExchange)getRawMessage(), getCommunicationPattern());
	}

	/**
	 * Check if this exchange is the first one for the current transition.
	 * @param exchange Exchange to check
	 * @param mep
	 * @return True if this exchange is the first one for the current transition.
	 */
	private boolean firstExchangeForTransition(MessageExchange exchange, String mep) {
		boolean firstExchangeForTransition = false;
		if (mep.equals(MEP_INONLY)) {
			debug(" MEP is IN_ONLY ");
			firstExchangeForTransition = exchange.getStatus().equals(ExchangeStatus.ACTIVE);
		} else if (mep.equals(MEP_INOUT)) {
			debug(" MEP is IN_OUT ");
			if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
				// I must check if this is the request or the response
				firstExchangeForTransition = (exchange.getRole() == MessageExchange.Role.CONSUMER);
			}
		} else if (mep.equals(MEP_ROBUSTINONLY)) {
			debug(" MEP is ROBUST_IN_ONLY ");
		} else if (mep.equals(MEP_INOPTIONALOUT)) {
			debug(" MEP is ROBUST IN_OPTIONAL_OUT ");
		} else {
			// Errore
        	log.error("firstExchangeForTransition. Mep type=UNDEFINED ");
		}
		return firstExchangeForTransition;
	}
	 
	//save the Message in output if there are the attachments 
	public void saveMessageOutForAttachments(TransitionInstance transitionInstance) {
		debug("saveMessageOutForAttachments: start...");
		if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)
				&& communicationPattern.equals(CommunicationPatterns.JBI_INOUT)) {
			int attachmentsOut = getNormalizedMessage().getAttachmentNames().size();
			debug("saveMessageOutForAttachments: attachmentsOut is" + attachmentsOut);
			if (attachmentsOut != 0) {
				transitionInstance.setTransitionDataOut(serializeNormalizedMessage());
				transitionInstance.setAttachmentOut(attachmentsOut);
			}
		}
	}
	
	public Set<String> getAttachmentNames(){
		return getNormalizedMessage().getAttachmentNames();
	}
	
	public void debug(Object o){
		if (log.isDebugEnabled()){
			log.debug(o);
		}
	}
}
